{{
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}}

{{
Improvements:
- Finetune accel/gyro and PID weights 
- Consider implementing servo deadband to minimize small movement
}}

CON
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000       '80 MHz

  RecordedFrame = logger#FramePitch

  'Controller states
  StateIdle = 1
  StateArmWait = 2
  StateArmed = 3
  StateFlying = 4 

OBJ
  IMU: "IMU"
  'debug: "SimpleDebug"
  logger: "Logger"
  fString: "FloatString"
  servoOut: "Servo32v3"
  servoIn: "ServoInput"
  fMath: "Float32"
  pitchPID: "PID"
  rollPID: "PID"
  utils: "Utils"
  quadMixer: "QuadMixer"
  cfg: "QuadConfig"

VAR

  'The following vars must be in sequence
  long xAxis 'float 
  long yAxis 'float
  long zAxis 'float
  long xRate 'float
  long yRate 'float
  long pitchAngle 'degrees, float
  long rollAngle 'degrees, float
  long pitchRate 'degress/sec, float
  long rollRate 'degrees/sec, float
  
  long gainIn 'ms, float
  long elevatorIn 'ms, float
  long aileronsIn 'ms, float
  long rudderIn 'ms, float
  long throttleIn 'ms, float
  
  long pitchCommand 'ms, float         
  long rollCommand 'ms, float         
  long yawCommand 'ms, float           
  long throttleCommand 'ms, float         
  
  long stabEnabled 'boolean
  long loopRunTime 'seconds, float
  long imuRunTime 'seconds, float
  
  long gainSetpoint 'Kp, float
  long pitchAngleSetpoint 'degrees, float
  long rollAngleSetpoint 'degrees, float

  long debug 'float
  long errorCode 'float
  long clock 'float
  long state 'float

  long motorFrontCommand 'ms, float
  long motorRearCommand 'ms, float
  long motorRightCommand 'ms, float     
  long motorLeftCommand 'ms, float   

DAT
  servoInPins           LONG cfg#PinGainIn, cfg#PinElevatorIn, cfg#PinAileronsIn, cfg#PinRudderIn, cfg#PinThrottleIn
  servoInPulseWidths    LONG 0, 0, 0, 0, 0
  servoInNumPins        LONG cfg#NumServoInputs

PUB main | startTimeCnt, remTimeCnt, loopPeriodCnt, n, tmp, runTimeCnt, tmpPitch, tmpRoll, tmpGain, tmpStabEnabled, imuRunning, clockCnt, stabLockout, servoInInitDelayCnt, pitchSetpointK, rollSetpointK, gainSetpointK, pitchSetpointO, rollSetpointO, loggerRTDelayCnt, tmpState, stateTimerCnt, throttleArmDelayCnt 
  'Initialize modules
  IMU.start(cfg#PinADCCS, cfg#PinADCDIO, cfg#PinADCCLK, cfg#IMUStartupDelay, cfg#IMUInitializeDelay, cfg#IMUWarmupDelay, cfg#IMULoopPeriodS)
  pitchPID.init(cfg#LoopPeriodS, cfg#PitchKp, cfg#PitchKi, cfg#PitchKd, cfg#PitchILimit)
  rollPID.init(cfg#LoopPeriodS, cfg#RollKp, cfg#RollKi, cfg#RollKd, cfg#RollILimit)
  quadMixer.init(cfg#PitchMixerGain, cfg#RollMixerGain, cfg#YawMixerGain, cfg#ServoMinValue, cfg#ServoMaxValue, cfg#ServoMidValue) 
  logger.start(@xAxis, RecordedFrame, true) ' Start in RT mode
  fMath.start                                              
  tmp := fMath.FRound(cfg#ServoMinValue)
  ServoIn.start(@servoInPins, servoInNumPins, @servoInPulseWidths)
  ServoOut.set(cfg#PinThrottleOut, tmp)
  ServoOut.set(cfg#PinElevatorOut, tmp)
  ServoOut.set(cfg#PinAileronsOut, tmp)
  ServoOut.set(cfg#PinRudderOut, tmp) 
  ServoOut.start

  'Initialize variables
  loopPeriodCnt := fMath.FRound( fMath.FMul(fMath.FFloat(clkfreq), cfg#LoopPeriodS) )
  servoInInitDelayCnt := fMath.FRound( fMath.FDiv(cfg#ServoInInitDelay, cfg#LoopPeriodS) )
  loggerRTDelayCnt := fMath.FRound( fMath.FDiv(cfg#LoggerRTDelay, cfg#LoopPeriodS) )
  throttleArmDelayCnt := fMath.FRound( fMath.FDiv(cfg#ThrottleArmDelay, cfg#LoopPeriodS) )
  pitchCommand := cfg#ServoMidValue
  yawCommand := cfg#ServoMidValue
  rollCommand := cfg#ServoMidValue
  throttleCommand := cfg#ServoMinValue
  gainIn := cfg#ServoMinValue   
  elevatorIn := cfg#ServoMidValue
  rudderIn := cfg#ServoMidValue
  aileronsIn := cfg#ServoMidValue
  throttleIn := cfg#ServoMinValue
  motorFrontCommand := cfg#ServoMinValue
  motorRearCommand := cfg#ServoMinValue
  motorRightCommand := cfg#ServoMinValue
  motorLeftCommand := cfg#ServoMinValue
  stabEnabled := false
  errorCode := 0.0
  clockCnt := 0
  stabLockout := false
  tmpState := StateIdle

  ' Pre calculate constants for scaling servo inputs to setpoints
  pitchSetpointK := fMath.FDiv(fMath.FSub(cfg#PitchMax, cfg#PitchMin), fMath.FSub(cfg#ServoMaxValue, cfg#ServoMinValue))
  rollSetpointK := fMath.FDiv(fMath.FSub(cfg#RollMax, cfg#RollMin), fMath.FSub(cfg#ServoMaxValue, cfg#ServoMinValue))
  gainSetpointK := fMath.FDiv(fMath.FSub(cfg#GainMax, cfg#GainMin), fMath.FSub(cfg#ServoMaxValue, cfg#ServoMidValue))
  pitchSetpointO := fMath.FSub(cfg#PitchMin, cfg#PitchSetpointOffset)
  rollSetpointO := fMath.FSub(cfg#RollMin, cfg#RollSetpointOffset) 
  
  'Main loop, executes every LoopPeriodS seconds 
  repeat
    
    'Timestamp start of loop
    startTimeCnt := cnt

    'Increment tick counter and update clock
    clockCnt := clockCnt + 1
    clock := fMath.FMul( fMath.FFloat(clockCnt), cfg#LoopPeriodS )

    'Calculate execution time of last loop
    loopRunTime := fMath.FDiv( fMath.FFloat(runTimeCnt), fMath.FFloat(clkfreq) )

    'Update and read IMU outputs
    xAxis := IMU.getXAccel
    yAxis := IMU.getYAccel
    zAxis := IMU.getZAccel
    xRate := IMU.getXRate
    yRate := IMU.getYRate
    pitchAngle := fMath.FAdd(IMU.getPitchAngle, cfg#PitchAngleOffset)
    rollAngle := fMath.FAdd(IMU.getRollAngle, cfg#RollAngleOffset)
    pitchRate := IMU.getPitchRate
    rollRate := IMU.getRollRate
    imuRunTime := IMU.getRunTime
    if IMU.getState == IMU#Running
      imuRunning := true
    else
      imuRunning := false

    'Check if the logger should exit RT mode and start recording in the EEPROM
    if (clockCnt == loggerRTDelayCnt) and (not logger.rtCmdReceived)
      logger.stop
      logger.start(@xAxis, RecordedFrame, false) ' Start in recording mode      
    
    'Read servo inputs, remove offset and limit range 
    repeat n from 0 to servoInNumPins-1
      tmp := servoInPulseWidths[n]
      if tmp > 0
        tmp := fMath.FFloat(tmp - cfg#ServoOffsetValueInt)
        tmp := fLimit(tmp, cfg#ServoMinValue, cfg#ServoMaxValue)
        longmove((@gainIn)+n*4, @tmp, 1)
      elseif clockCnt > servoInInitDelayCnt
        errorCode := cfg#ErrorServoIn

    'Update controller state
    if tmpState == StateIdle
      'Wait for throttle to go below threshold
      if fMath.FCmp(throttleIn, cfg#ThrottleArmThreshold) < 0   
        tmpState := StateArmWait
        stateTimerCnt := clockCnt + throttleArmDelayCnt 
    elseif tmpState == StateArmWait
      'Make sure that throttle stays below threshold for the specified delay before arming motors
      if fMath.FCmp(throttleIn, cfg#ThrottleArmThreshold) > 0
        tmpState := StateIdle
      elseif clockCnt == stateTimerCnt
        tmpState := StateArmed
    elseif tmpState == StateArmed
      'When the throttle goes above the threshold transision to Flying state and reset PID loops
      if fMath.FCmp(throttleIn, cfg#ThrottleArmThreshold) > 0
        tmpState := StateFlying
        pitchPID.reset
        rollPID.reset
    elseif tmpState == StateFlying
      if fMath.FCmp(throttleIn, cfg#ThrottleArmThreshold) < 0
        tmpState := StateArmed
    else
      tmpState := StateIdle
    state := fMath.FFloat(tmpState)                         

    'Calculate pitch angle setpoint (elevator) 
    tmpPitch := fMath.FSub(elevatorIn, cfg#ServoMinValue)
    tmpPitch := fMath.FMul(tmpPitch, pitchSetpointK)
    tmpPitch := fMath.FAdd(tmpPitch, pitchSetpointO)   
    if cfg#ElevatorReverse
      pitchAngleSetpoint := fMath.FNeg(tmpPitch)
    else
      pitchAngleSetpoint := tmpPitch

    'Calculate roll angle setpoint (ailerons)     
    tmpRoll := fMath.FSub(aileronsIn, cfg#ServoMinValue)
    tmpRoll := fMath.FMul(tmpRoll, rollSetpointK)
    tmpRoll := fMath.FAdd(tmpRoll, rollSetpointO)  
    if cfg#AileronsReverse
      rollAngleSetpoint := fMath.FNeg(tmpRoll)
    else
      rollAngleSetpoint := tmpRoll
      
    'Calculate gain setpoint 
    if fMath.fCmp(gainIn, cfg#ServoMidValue) > 0
      tmpStabEnabled := true
      tmpGain := fMath.FSub(gainIn, cfg#ServoMidValue)
      tmpGain := fMath.FMul(tmpGain, gainSetpointK)
      gainSetpoint := fMath.FAdd(tmpGain, cfg#GainMin)
    else
      tmpStabEnabled := false  
      gainSetpoint := 0.0 

    'Check IMU output for validity and determine if stabilization should be enabled
    if fMath.FCmp(pitchAngle, cfg#AngleMin) < 0 or fMath.FCmp(pitchAngle, cfg#AngleMax) > 0 or fMath.FCmp(rollAngle, cfg#AngleMin) < 0 or fMath.FCmp(rollAngle, cfg#AngleMax) > 0
      stabEnabled := false
      stabLockout := true
      errorCode := cfg#ErrorInvalidAngles
    else
      if not tmpStabEnabled  'Gain negative so stab is disabled, also reset lockout and pid loops
        stabEnabled := false
        stabLockout := false
        pitchPID.reset
        rollPID.reset
      else
        if imuRunning and (not stabLockout) 'Enable stab only if gain is possitive and not in lockout mode
          stabEnabled := true
        else
          stabEnabled := false

    'Update logger status
    logger.enable(stabEnabled)
    
    'Calculate new servo outputs (PID loop if stabilization is enabled, or pass through otherwise)
    if stabEnabled
      'Update PID gains
      pitchPID.setGain( fMath.FAdd(cfg#PitchKp, gainSetpoint) )   
      rollPID.setGain( fMath.FAdd(cfg#RollKp, gainSetpoint) )  
      'Update PID loops
      fMath.stop
      tmpPitch := pitchPID.updateR(pitchAngleSetpoint, pitchAngle, pitchRate)
      tmpRoll := rollPID.updateR(rollAngleSetpoint, rollAngle, rollRate)
      fMath.start
      'Reverse output if required
      if cfg#ElevatorReverse
        tmpPitch := fMath.FNeg(tmpPitch)
      if cfg#AileronsReverse
        tmpRoll := fMath.FNeg(tmpRoll)      
      'Offset from servo mid point
      tmpPitch := fMath.FAdd(tmpPitch, cfg#ServoMidValue)
      tmpRoll := fMath.FAdd(tmpRoll, cfg#ServoMidValue)  
      'Limit outputs
      tmpPitch := fLimit(tmpPitch, cfg#ServoMinValue, cfg#ServoMaxValue)
      tmpRoll := fLimit(tmpRoll, cfg#ServoMinValue, cfg#ServoMaxValue) 
      'Set channel final values
      if cfg#PitchControlEnable
        pitchCommand := tmpPitch
      else
        pitchCommand := elevatorIn
      if cfg#RollControlEnable  
        rollCommand := tmpRoll
      else
        rollCommand := aileronsIn   
      throttleCommand := throttleIn
      yawCommand := rudderIn 
    else
      throttleCommand := throttleIn
      pitchCommand := elevatorIn
      yawCommand := rudderIn
      rollCommand := aileronsIn

    'Calculate and set outputs depending on mode
    if cfg#QuadMode
      'Calculate motor outputs
      if cfg#Passthrough
        motorFrontCommand := throttleCommand
        motorRearCommand := throttleCommand  
        motorRightCommand := throttleCommand  
        motorLeftCommand := throttleCommand
      elseif tmpState == StateArmed or tmpState == StateFlying
        fMath.stop
        quadMixer.update(throttleCommand, pitchCommand, rollCommand, yawCommand)
        fMath.start 
        motorFrontCommand := quadMixer.getMotorFront
        motorRearCommand := quadMixer.getMotorRear  
        motorRightCommand := quadMixer.getMotorRight  
        motorLeftCommand := quadMixer.getMotorLeft
      else
        motorFrontCommand := 0.0
        motorRearCommand := 0.0  
        motorRightCommand := 0.0  
        motorLeftCommand := 0.0          
      'Set outputs (add offset back in)
      if cfg#EnableOutputs
        ServoOut.set(cfg#PinThrottleOut, fMath.FRound(motorFrontCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinElevatorOut, fMath.FRound(motorRearCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinRudderOut, fMath.FRound(motorRightCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinAileronsOut, fMath.FRound(motorLeftCommand) + cfg#ServoOffsetValueInt)       
    else
      'Limit roll and pitch to end points
      tmpPitch := fLimit(pitchCommand, cfg#ElevatorEndPointL, cfg#ElevatorEndPointH)
      tmpRoll := fLimit(rollCommand, cfg#AileronsEndPointL, cfg#AileronsEndPointH) 
      'Set outputs (add offset back in)
      if cfg#EnableOutputs      
        ServoOut.set(cfg#PinThrottleOut, fMath.FRound(throttleCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinElevatorOut, fMath.FRound(pitchCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinRudderOut, fMath.FRound(yawCommand) + cfg#ServoOffsetValueInt)
        ServoOut.set(cfg#PinAileronsOut, fMath.FRound(rollCommand) + cfg#ServoOffsetValueInt)  
  
    'Wait for the remaining of the update period
    runTimeCnt := cnt - startTimeCnt
    remTimeCnt := loopPeriodCnt - runTimeCnt
    if (remTimeCnt > 0) and (remTimeCnt < loopPeriodCnt)
      waitcnt(remTimeCnt + cnt)
    else
      errorCode := cfg#ErrorLoopOverrun

PRI fLimit(value, minValue, maxValue)
  if fMath.FCmp(value, minValue) < 0
    return minValue
  elseif fMath.FCmp(value, maxValue) > 0
    return maxValue
  else
    return value            
    