{{
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}}

OBJ
  uart : "Simple_Serial"
  i2c : "Basic_I2C_Driver" 
  utils : "Utils"
  fString : "FloatString"

CON
  UARTRxPin = 31
  UARTTxPin = 30
  EEPROMSCLPin = 8
  EEPROMSDAPin = 9

  EEPROMPageSize = 64
  EEPROMMemorySize = 32768
  EEPROMAddr = $A0
  
  UARTBaud = 19200

  FrameRT = 0
  FrameRTNumChannels = 32
  FrameRTLoopPeriod = 100 'ms

  FrameAngle = 3
  FrameAngleNumChannels = 2
  FrameAngleLoopPeriod = 10 'ms

  FramePitch = 1
  FramePitchNumChannels = 4
  FramePitchLoopPeriod = 100 'ms

  FrameRoll = 2
  FrameRollNumChannels = 4
  FrameRollLoopPeriod = 100 'ms

  FrameIMU = 4
  FrameIMUNumChannels = 8
  FrameIMULoopPeriod = 20 'ms    
  
DAT
  FramePitchChannels    LONG    5, 14, 20, 21 'Index to IMUStab.VAR section
  FrameRollChannels     LONG    6, 15, 20, 22 'Index to IMUStab.VAR section
  FrameIMUChannels      LONG    0,1,2,3,4,5,10,11 'Index to IMUStab.VAR section
  FrameAngleChannels    LONG    5, 6 'Index to IMUStab.VAR section   

VAR
  long stack[100]
  long frameType
  long dataPtr
  long recEnabled
  long eepromBuffer[EEPROMPageSize/4]
  long eepromBufferIdx
  long eepromAddrIdx
  long loopPeriodCnt
  long channels
  long numChannels
  long cid
  long rtMode
  long cmdReceived

PUB start(_dataPtr, _frameType, _rtMode)
  frameType := _frameType
  dataPtr := _dataPtr
  rtMode := _rtMode
  recEnabled := false
  cmdReceived := false
  eepromBufferIdx := 0
  eepromAddrIdx := 0

  cid := cognew(main, @stack)

PUB stop
  cogstop(cid)
  
PUB enable(state)
  recEnabled := state

PUB rtCmdReceived
  return cmdReceived  

PRI main | cmd, param
  ' If in RT mode, open UART and check for terminal commands
  if rtMode
    uart.init(UARTRxPin, UARTTxPin, UARTBaud)
    repeat
      cmd := uart.rx
      'Execute command 
      if cmd == "r"
        cmdReceived := true
        loopPeriodCnt := (clkfreq / 1000) * FrameRTLoopPeriod
        frameType := FrameRT
        numChannels := FrameRTNumChannels
        realTimeDisplay(true) 'does not return!
      if cmd == "a"
        cmdReceived := true
        loopPeriodCnt := (clkfreq / 1000) * FrameAngleLoopPeriod
        frameType := FrameAngle
        channels := @FrameAngleChannels
        numChannels := FrameAngleNumChannels
        realTimeDisplay(false) 'does not return!          
      elseif cmd == "d"
        cmdReceived := true
        dumpMemory
      elseif cmd == "e"
        cmdReceived := true
        eraseMemory
  else
    'Setup for selected frame type
    if frameType == FramePitch
      'Calculate loop period counter
      loopPeriodCnt := (clkfreq / 1000) * FramePitchLoopPeriod
      'Setup channels
      channels := @FramePitchChannels
      numChannels := FramePitchNumChannels
    elseif frameType == FrameRoll
      'Calculate loop period counter
      loopPeriodCnt := (clkfreq / 1000) * FrameRollLoopPeriod
      'Setup channels
      channels := @FrameRollChannels
      numChannels := FrameRollNumChannels
    elseif frameType == FrameIMU
      'Calculate loop period counter
      loopPeriodCnt := (clkfreq / 1000) * FrameIMULoopPeriod
      'Setup channels
      channels := @FrameIMUChannels
      numChannels := FrameIMUNumChannels
   'Start recorder
   memoryRecorder 'does not return

PRI memoryRecorder | remTimeCnt, startTimeCnt, n, chIdx, stat, tmp
  'Initialize I2C bus
  i2c.Initialize(EEPROMSCLPin)

  'Recording loop
  repeat until eepromAddrIdx == EEPROMMemorySize
    'Timestamp start of loop
    startTimeCnt := cnt

    'Check if recording is enabled
    if recEnabled
      'Copy data to buffer
      repeat n from 0 to numChannels-1
        longmove(@chIdx, channels+n*4, 1)
        longmove(@eepromBuffer + (eepromBufferIdx*4), dataPtr+(chIdx*4), 1)
        eepromBufferIdx := eepromBufferIdx + 1
                
      'If buffer is full send data to EEPROM
      if eepromBufferIdx == EEPROMPageSize/4             
        stat:=i2c.WritePage(EEPROMSCLPin, EEPROMAddr, eepromAddrIdx, @eepromBuffer, EEPROMPageSize)  
        eepromBufferIdx := 0
        eepromAddrIdx := eepromAddrIdx + EEPROMPageSize
      
    'Wait for the remaining of the update period 
    remTimeCnt := loopPeriodCnt - (cnt - startTimeCnt)
    if (remTimeCnt > 0) and (remTimeCnt < loopPeriodCnt) 
      waitcnt(cnt + remTimeCnt)
 
PRI realTimeDisplay(sendAll) | startTimeCnt, remTimeCnt, tmpChannel, n, chIdx
  repeat
    'Timestamp start of loop
    startTimeCnt := cnt

    'Send sync
    str(String("imud"))

    'Send frame id
    uart.tx(frameType)
    
    'Send out each channel (as bytes)
    repeat n from 0 to numChannels-1
      if sendAll
        longmove(@tmpChannel, dataPtr+(4*n), 1)
      else  
        longmove(@chIdx, channels+n*4, 1)
        longmove(@tmpChannel, dataPtr+(chIdx*4), 1)
      uart.tx(tmpChannel & $FF)
      uart.tx(tmpChannel>>8 & $FF)
      uart.tx(tmpChannel>>16 & $FF)
      uart.tx(tmpChannel>>24 & $FF)
          
    'Wait for the remaining of the update period
    remTimeCnt := loopPeriodCnt - (cnt - startTimeCnt)
    if (remTimeCnt > 0) and (remTimeCnt < loopPeriodCnt) 
      waitcnt(cnt + remTimeCnt) 

PRI dumpMemory | n, tmpChannel, stat
  'Initialize I2C bus
  i2c.Initialize(EEPROMSCLPin)

  'Dump memory page by page
  repeat until eepromAddrIdx == EEPROMMemorySize
    'Read page
    stat := i2c.ReadPage(EEPROMSCLPin, EEPROMAddr, eepromAddrIdx, @eepromBuffer, EEPROMPageSize)    
    eepromAddrIdx := eepromAddrIdx + EEPROMPageSize 
    repeat eepromBufferIdx from 0 to (EEPROMPageSize/4)-1 step numChannels
      'Send sync
      str(String("imud"))    
      'Send out frame type
      uart.tx(frameType)
      'Write one frame per line
      repeat n from 0 to numChannels-1
        longmove(@tmpChannel, @eepromBuffer+(4*eepromBufferIdx)+(4*n), 1)
        uart.tx(tmpChannel & $FF)
        uart.tx(tmpChannel>>8 & $FF)
        uart.tx(tmpChannel>>16 & $FF)
        uart.tx(tmpChannel>>24 & $FF)

  'Signal end of operation
  str(String("imud"))
  uart.tx(-1)                 


PRI eraseMemory
  'Signal start of operation
  str(String("imud"))
  uart.tx(0)  
  'Fill buffer with zeros
  repeat eepromBufferIdx from 0 to (EEPROMPageSize/4)-1
    eepromBuffer[eepromBufferIdx] := 0
  'Write buffer  
  repeat eepromAddrIdx from 0 to EEPROMMemorySize-1 step EEPROMPageSize
    i2c.WritePage(EEPROMSCLPin, EEPROMAddr, eepromAddrIdx, @eepromBuffer, EEPROMPageSize)
    repeat while i2c.WriteWait(EEPROMSCLPin, EEPROMAddr, eepromAddrIdx)
  'Signal end of operation
  str(String("imud"))
  uart.tx(-1)        

PRI str(stringPtr) | i
'' Print a zero-terminated string
  repeat i from 0 to strsize(stringPtr) - 1
    putc(byte[stringPtr][i])


PRI dec(value) | i, z
'' Print a signed decimal number
  if value < 0
    -value
    putc("-")

  i := 1_000_000_000
  z~

  repeat 10
    if value => i
      putc(value / i + "0")
      value //= i
      z~~
    elseif z or i == 1
      putc("0")
    i /= 10

PUB putc(txbyte)
  uart.tx(txbyte)                  