{{
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}}

OBJ
  fMath: "Float32"

VAR
  long periodS
  long Kp
  long Ki
  long Kd
  long integral
  long previousError
  long integralLimit
  
PUB init(_periodS, _Kp, _Ki, _Kd, _integralLimit)
  periodS := _periodS
  Kp := _Kp
  Ki := _Ki
  Kd := _Kd
  integralLimit := _integralLimit
  reset

PUB reset
  integral := 0.0
  previousError := 0.0

PUB setGain(_Kp)
  Kp := _Kp

PUB updateR(setpoint, actual, rate): output | error, derivative, pTerm, iTerm, dTerm, iLimit
  ' Start math
  fMath.start 
  ' Calculate error, integral, derivative
  error := fMath.FSub(setpoint, actual)
  integral := fMath.FAdd(integral, fMath.FMul(error, periodS)) 
  'derivative := fMath.FDiv(fMath.FSub(error,previousError), periodS)  
  derivative := rate   
  previousError := error
  ' Calculate integral limit for current Ki
  iLimit := fMath.FDiv(integralLimit, Ki)
  ' Limit integral
  if fMath.FCmp(integral, fMath.FNeg(iLimit)) < 0
    integral := fMath.FNeg(iLimit)
  elseif fMath.FCmp(integral, iLimit) > 0
    integral := iLimit
  ' Multiply with constants
  pTerm := fMath.FMul(Kp, error)
  iTerm := fMath.FMul(Ki, integral) 
  dTerm := fMath.FMul(Kd, derivative)    
  ' Calculate output
  output := fMath.FAdd( pTerm, iTerm ) 
  output := fMath.FAdd( output, dTerm )
  ' Done
  fMath.stop
  return output