{{
    Copyright 2009 Vassilis Varveropoulos (vassilis@vrhome.net)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
}}

OBJ
  fMath: "Float32"

VAR
  long motorFront
  long motorRear
  long motorRight
  long motorLeft

  long pitchMixerGain
  long rollMixerGain
  long yawMixerGain
  long cmdMin
  long cmdMax
  long cmdNeutral
  long throttleMin
  long throttleMax

PUB init(_pitchMixerGain, _rollMixerGain, _yawMixerGain, _cmdMin, _cmdMax, _cmdNeutral) | tmpOffset
  fMath.start 
  longmove(@pitchMixerGain, @_pitchMixerGain, 6)
  tmpOffset := fMath.FAdd(fMath.FMax(pitchMixerGain, rollMixerGain), yawMixerGain)
  tmpOffset := fMath.FDiv(tmpOffset, 4.0) 
  tmpOffset := fMath.FMul(tmpOffset, fMath.FSub(cmdMax, cmdMin))
  throttleMin := fMath.FAdd(cmdMin, tmpOffset)
  throttleMax := fMath.FSub(cmdMax, tmpOffset)
  fMath.stop

PUB getMotorFront
  return motorFront

PUB getMotorRear
  return motorRear

PUB getMotorRight
  return motorRight

PUB getMotorLeft
  return motorLeft        

PUB update(throttleCommand, pitchCommand, rollCommand, yawCommand)
  'Start math
  fMath.start 
  'If the throttle is below the minimum limit then mixing is disabled
  if fMath.FCmp(throttleCommand, throttleMin) < 0
    motorFront := throttleCommand
    motorRear := throttleCommand
    motorRight := throttleCommand
    motorLeft := throttleCommand
  else
    'Limit the throttle between the minimum and maximum 
    throttleCommand := fLimit(throttleCommand, throttleMin, throttleMax)
    'Move neutral of commands to zero
    pitchCommand := fMath.FSub(pitchCommand, cmdNeutral)
    rollCommand := fMath.FSub(rollCommand, cmdNeutral)
    yawCommand := fMath.FSub(yawCommand, cmdNeutral)   
    'Apply gain on pitch, roll and yaw commands
    pitchCommand := fMath.FMul(pitchCommand, pitchMixerGain)
    rollCommand := fMath.FMul(rollCommand, rollMixerGain)
    yawCommand := fMath.FMul(yawCommand, yawMixerGain)  
    'Calculate motor outputs
    motorFront := fMath.FAdd(fMath.FAdd( throttleCommand, pitchCommand ), yawCommand)  ' throttle + pitch + yaw
    motorRear := fMath.FAdd(fMath.FSub( throttleCommand, pitchCommand ), yawCommand)   ' throttle - pitch + yaw
    motorRight := fMath.FSub(fMath.FAdd( throttleCommand, rollCommand ), yawCommand)   ' throttle + roll - yaw
    motorLeft := fMath.FSub(fMath.FSub( throttleCommand, rollCommand ), yawCommand)    ' throttle - roll - yaw
    'Limit motor outputs
    motorFront := fLimit(motorFront, cmdMin, cmdMax)
    motorRear := fLimit(motorRear, cmdMin, cmdMax)
    motorRight := fLimit(motorRight, cmdMin, cmdMax)
    motorLeft := fLimit(motorLeft, cmdMin, cmdMax)
  'Done
  fMath.stop  

PRI fLimit(value, minValue, maxValue)
  if fMath.FCmp(value, minValue) < 0
    return minValue
  elseif fMath.FCmp(value, maxValue) > 0
    return maxValue
  else
    return value  